import pytest  # type: ignore[import]

# noinspection PyUnresolvedReferences
from nhs_aws_helpers.fixtures import *  # noqa: F403
from tests.utils import temp_env_vars


@pytest.fixture(scope="session", autouse=True)
def autouse():
    with temp_env_vars(AWS_ENDPOINT_URL="http://localhost:4566"):
        yield



from io import SEEK_CUR, SEEK_END, BytesIO
from time import time
from typing import cast
from uuid import uuid4

import pytest
from mypy_boto3_s3.service_resource import Bucket, Object

from nhs_aws_helpers import s3_object
from nhs_aws_helpers.common import MiB
from nhs_aws_helpers.s3_object_reader import S3ObjectReader
from nhs_aws_helpers.s3_object_writer import S3ObjectWriter

_DEFAULT_BUFFER_SIZE = 200

_TEXT = [
    "It bégan with the forging of the Great Rings.\r\n",
    "Threé were given to the Elves, immortal, wisest and fairest of all beings.\n",
    "Seven to the Dwarf-Lords, great miners and craftsmen of the mountain halls.\r\n",
    "And nine, nine rings were gifted to the race of Men, who above all else desire power.\r\n",
    "For within these rings was bound the strength and the will to govern each race.\r\n",
    "But théy were all of them deceived, for another ring was made.\r\n",
    "Déép in the land of Mordor, in thé Fires of Mount Doom, thé Dark Lord Sauron forged a master "
    "ring in secret, and into this ring he poured his cruelty, his malice and his will to dominaté "
    "all life.\n",
    "Oné ring to rule them all.",
]
_S3_TEXT_BODY = "".join(_TEXT)
_TEXT_BODY_LEN = len(_S3_TEXT_BODY)

_ASCII_TEXT = [
    "It began with the forging of the Great Rings.\r\n",
    "Three were given to the Elves, immortal, wisest and fairest of all beings.\n",
    "Seven to the Dwarf-Lords, great miners and craftsmen of the mountain halls.\r\n",
    "And nine, nine rings were gifted to the race of Men, who above all else desire power.\r\n",
    "For within these rings was bound the strength and the will to govern each race.\r\n",
    "But they were all of them deceived, for another ring was made.\r\n",
    "Deep in the land of Mordor, in the Fires of Mount Doom, the Dark Lord Sauron forged a master "
    "ring in secret, and into this ring he poured his cruelty, his malice and his will to dominate "
    "all life.\n",
    "One ring to rule them all.",
]
_ASCII_S3_TEXT_BODY = "".join(_ASCII_TEXT)
_ASCII_TEXT_BODY_LEN = len(_ASCII_S3_TEXT_BODY)

_NON_ASCII_TEXT = [
    "It begдn with the forging of the Great Rings.\r\n",
    "Three were given to the Elves, immortдl, wisest and fairest of дll beings.\n",
    "Seven to the Dwarf-Lords, great miners and craftsmen of the mountaiи halls.\r\n",
    "And nine, nine rings were gifted to the race of Men, who above all else desiлe power.\r\n",
    "For witлin these rings was bound the strength and the will to govern each race.\r\n",
    "But they were all of them deceived, for another ring was made.\r\n",
    "Deep in the land of Mordor, in the Fires of Mount Doom, the Dark Lord Sauлon forged a master "
    "ring in secret, and into this ring he poured his cruelty, his malice and his will to dominate "
    "all life.\n",
    "One ring to rule them all. Владимир Путин",
]
_NON_ASCII_S3_TEXT_BODY = "".join(_NON_ASCII_TEXT)
_NON_ASCII_TEXT_BODY_LEN = len(_NON_ASCII_S3_TEXT_BODY)
_NON_ASCII_TEXT_BODY_LEN_ENCODED = len(_NON_ASCII_S3_TEXT_BODY.encode("utf-8"))

_BINARY = b"Example bytes"


class MockS3Object:
    def __init__(self, bytes_io):
        self.bytes_io = bytes_io

    def get(self):
        return {"Body": self.bytes_io}

    def shh(self):
        return self.__dict__


def s3_reader_text():
    bytes_io = BytesIO(_S3_TEXT_BODY.encode("utf-8"))

    s3_io = S3ObjectReader(
        s3_obj=cast(Object, MockS3Object(bytes_io)), buffer_size=_DEFAULT_BUFFER_SIZE, encoding="utf-8"
    )

    return s3_io


def ascii_s3_reader_text():
    bytes_io = BytesIO(_ASCII_S3_TEXT_BODY.encode("utf-8"))

    s3_io = S3ObjectReader(
        s3_obj=cast(Object, MockS3Object(bytes_io)), buffer_size=_DEFAULT_BUFFER_SIZE, encoding="utf-8"
    )

    return s3_io


def non_ascii_s3_reader_text():
    bytes_io = BytesIO(_NON_ASCII_S3_TEXT_BODY.encode("utf-8"))

    s3_io = S3ObjectReader(
        s3_obj=cast(Object, MockS3Object(bytes_io)), buffer_size=_DEFAULT_BUFFER_SIZE, encoding="utf-8"
    )

    return s3_io


def s3_reader_binary():
    bytes_io = BytesIO(_BINARY)

    s3_io = S3ObjectReader(s3_obj=cast(Object, MockS3Object(bytes_io)), buffer_size=_DEFAULT_BUFFER_SIZE)

    return s3_io


def s3_writer(bucket: Bucket, key: str):
    return S3ObjectWriter(s3_object(bucket.name, key), encoding="utf-8")


def s3_writer_binary(bucket: Bucket, key: str):
    return S3ObjectWriter(s3_object(bucket.name, key))


def test_read_all():
    with s3_reader_text() as reader:
        body = reader.read()

    assert body == _S3_TEXT_BODY


def test_read_n() -> None:
    size = 25

    chunks = []

    with s3_reader_text() as reader:
        chunk = reader.read(size)
        pos = 0
        assert len(reader._buffer.encode("utf-8")) == _DEFAULT_BUFFER_SIZE  # type: ignore[union-attr]

        while chunk:
            chunks.append(chunk)
            assert _S3_TEXT_BODY[pos : pos + len(chunk)] == chunk
            pos += len(chunk)
            chunk = reader.read(size)

    assert "".join(chunks) == _S3_TEXT_BODY  # type: ignore[arg-type]


def test_read_negative() -> None:
    with s3_reader_text() as reader:
        first_chunk = reader.read(300)

        neg_chunk = reader.read(-33)

        assert _S3_TEXT_BODY.encode("utf-8")[0:300].decode("utf-8") == first_chunk
        assert first_chunk[-33:] == neg_chunk
        assert reader._bytes_read == 298
        assert reader._position == 265


def test_read_negative_all() -> None:
    read_size = -57

    with s3_reader_text() as reader:
        _ = reader.read()
        chunks: list[str] = []
        chunk = reader.read(read_size)
        while chunk:
            chunks.insert(0, chunk)
            chunk = reader.read(read_size)
        joined = "".join(chunks)
        assert joined == _S3_TEXT_BODY


def test_seek_back_and_forth() -> None:
    with non_ascii_s3_reader_text() as reader:
        pos = reader.seek(57)

        assert pos == 57
        assert reader._position == 57
        assert reader._bytes_read <= _DEFAULT_BUFFER_SIZE

        pos = reader.seek(57)

        assert pos == 57
        assert reader._position == 57
        assert reader._bytes_read <= _DEFAULT_BUFFER_SIZE

        new_pos = _DEFAULT_BUFFER_SIZE + 57
        pos = reader.seek(new_pos)
        assert new_pos == pos
        assert new_pos == reader._position
        assert reader._bytes_read <= 2 * _DEFAULT_BUFFER_SIZE

        pos = reader.seek(_NON_ASCII_TEXT_BODY_LEN_ENCODED + 1)

        assert pos == _NON_ASCII_TEXT_BODY_LEN

        pos = reader.seek(100, SEEK_END)

        assert pos == _NON_ASCII_TEXT_BODY_LEN - 100


def test_seek_relative() -> None:
    pos = 0

    with non_ascii_s3_reader_text() as reader:
        for i in range(7):
            pos = reader.seek(100, SEEK_CUR)

            assert pos == min((i + 1) * 100, _NON_ASCII_TEXT_BODY_LEN)

        assert pos == _NON_ASCII_TEXT_BODY_LEN

        for i in range(7):
            pos = reader.seek(-100, SEEK_CUR)

            assert pos == max(0, _NON_ASCII_TEXT_BODY_LEN - ((i + 1) * 100))

        assert pos == 0

        read_all = reader.read()

        assert read_all == _NON_ASCII_S3_TEXT_BODY


def test_read_line() -> None:
    with s3_reader_text() as reader:
        line = reader.readline()

        assert line == _TEXT[0]

        line = reader.readline()

        assert line == _TEXT[1]

        for _ in range(len(_TEXT) - 2):
            line = reader.readline()

        assert line == _TEXT[-1]


def test_read_line_on_binary_object():
    with pytest.raises(ValueError, match="readline only works if _encoding was defined"), s3_reader_binary() as reader:
        reader.readline()


def test_read_lines() -> None:
    with s3_reader_text() as reader:
        line = reader.readline()

        assert (_TEXT[0]) == line

        lines = list(reader.readlines())

        assert (_TEXT[1:]) == lines


def test_read_lines_on_binary_object():
    with pytest.raises(ValueError, match="readlines only works if _encoding was defined"):
        list(s3_reader_binary().readlines())


def test_iter() -> None:
    with s3_reader_text() as reader:
        for i, a_line in enumerate(reader):
            assert _TEXT[i].strip() == a_line


def test_write_binary(temp_s3_bucket) -> None:
    key = f"testkey/{uuid4().hex}"
    with s3_writer_binary(temp_s3_bucket, key) as writer:
        writer.write(_BINARY)
        writer.close()

    read = temp_s3_bucket.Object(key).get()["Body"].read()

    assert read == _BINARY


def test_write_binary_then_nothing(temp_s3_bucket) -> None:
    key = f"testkey/{uuid4().hex}"
    with s3_writer_binary(temp_s3_bucket, key) as writer:
        writer.write(_BINARY)
        writer.write(b"")
        writer.close()

    read = temp_s3_bucket.Object(key).get()["Body"].read()

    assert read == _BINARY


def test_write_string(temp_s3_bucket) -> None:
    key = f"testkey/{uuid4().hex}"
    with s3_writer(temp_s3_bucket, key) as writer:
        writer.writelines(_S3_TEXT_BODY)
        writer.close()

    read = temp_s3_bucket.Object(key).get()["Body"].read()
    decoded = read.decode("utf-8")
    assert decoded == _S3_TEXT_BODY


def test_write_lines_string(temp_s3_bucket) -> None:
    key = f"testkey/{uuid4().hex}"
    with s3_writer(temp_s3_bucket, key) as writer:
        writer.writelines(_TEXT)
        writer.close()

    read = temp_s3_bucket.Object(key).get()["Body"].read()
    decoded = read.decode("utf-8")
    assert decoded == _S3_TEXT_BODY


def test_upload_large_multipart_file(temp_s3_bucket) -> None:
    bytes_to_write = 20 * MiB
    block = b"A" * (4 * MiB)
    key = f"testkey/{uuid4().hex}"
    start = time()
    with s3_writer_binary(temp_s3_bucket, key) as writer:
        bytes_written = 0
        while bytes_written < bytes_to_write:
            bytes_written += writer.write(block)
        writer.flush()
        etag = writer.s3_obj.e_tag
        assert etag.endswith('-3"')
    elapsed = time() - start
    assert elapsed < 2

    with S3ObjectReader(s3_object(temp_s3_bucket.name, key)) as reader:
        bytes_read = reader.read()
    assert len(bytes_read) == bytes_to_write



from collections.abc import Iterable, Iterator, Mapping
from concurrent.futures import Future, ThreadPoolExecutor
from typing import (
    IO,
    Any,
    AnyStr,
    Final,
    Optional,
    Union,
)

from mypy_boto3_s3.service_resource import MultipartUpload, Object

from nhs_aws_helpers.common import MiB


class S3ObjectWriter(IO):
    """provides in-memory buffer to write to S3 objects"""

    MIN_PART_SIZE: Final[int] = 5 * MiB
    DEFAULT_BUFFER_SIZE: Final[int] = 8 * MiB

    def __init__(
        self,
        s3_obj: Object,
        encoding: Optional[str] = None,
        buffer_size: int = DEFAULT_BUFFER_SIZE,
        name: str = "",
        create_object_args: Optional[Mapping[str, Any]] = None,
        max_multipart_concurrency: int = 12,
    ) -> None:
        buffer_size = max(buffer_size, self.MIN_PART_SIZE)
        self._s3_obj = s3_obj
        self._closed = False
        self._encoding = encoding
        self._buffer = b""
        self._bytes_written, self._position = (0, 0)
        self._parts: list[dict[str, int]] = []
        self._upload_tasks: list[Future] = []
        self._thread_pool: Optional[ThreadPoolExecutor] = None
        self._upload_errors: list[Exception] = []
        self._buffer_size = buffer_size
        self._multipart_upload: Optional[MultipartUpload] = None
        self._max_multipart_concurrency = max_multipart_concurrency
        self._name = name
        self._create_object_args = dict(kv for kv in create_object_args.items()) if create_object_args else {}

    @property
    def s3_obj(self) -> Object:
        return self._s3_obj

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        failed = exc_type is not None
        self._close(failed)
        return not failed

    def read(self, n: int = ...) -> Union[str, bytes]:
        raise NotImplementedError

    def readline(self, limit: int = ...) -> Union[str, bytes]:
        raise NotImplementedError

    def readlines(self, hint: int = ...) -> list[Union[str, bytes]]:
        raise NotImplementedError

    def seek(self, offset: int, whence: int = ...) -> int:
        raise NotImplementedError

    def __next__(self) -> Union[str, bytes]:
        raise NotImplementedError

    def __iter__(self) -> Iterator[Union[str, bytes]]:
        raise NotImplementedError

    @property
    def mode(self) -> str:
        return "w" if self._encoding else "wb"

    @property
    def name(self) -> str:
        return self._name

    @property
    def closed(self):
        return self._closed

    def fileno(self):
        return self._s3_obj.key

    def flush(self):
        self._maybe_write()

    def readable(self):
        return False

    def readinto(self, byte_array: bytearray) -> int:
        raise NotImplementedError

    def peek(self, num: Optional[int] = None) -> Union[str, bytes]:
        raise NotImplementedError

    def seekable(self) -> bool:
        return False

    def isatty(self) -> bool:
        return False

    def tell(self):
        return self._position

    def truncate(self, size=None):
        raise NotImplementedError

    def writable(self) -> bool:
        return True

    def writelines(self, lines: Iterable[Any]) -> None:
        for arg in lines:
            self.write(arg)

    def write(self, data: Any) -> int:
        """Write the given bytes (binary string) to the S3 file.
        There's buffering happening under the covers, so this may not actually
        do any HTTP transfer right away."""

        encoded = self._maybe_encode(data)
        self._buffer = b"".join([self._buffer, encoded]) if self._buffer else encoded  # type: ignore[list-item]

        self._position += len(encoded)

        while self.tell() >= self._buffer_size:
            self._start_next_part_upload()

        return len(encoded)

    def _raise_if_errored(self):
        if len(self._upload_errors) < 1:
            return

        raise RuntimeError from self._upload_errors[0]

    def _start_next_part_upload(self):
        self._raise_if_errored()

        if self._multipart_upload is None:
            args = self._create_object_args or {}
            self._multipart_upload = self._s3_obj.initiate_multipart_upload(**args)
            self._thread_pool = ThreadPoolExecutor(self._max_multipart_concurrency)

        part_data = self._buffer[: self._buffer_size]
        part_num = len(self._upload_tasks) + 1

        assert self._thread_pool
        self._upload_tasks.append(self._thread_pool.submit(self._upload_part, part_num, part_data))

        self._buffer = self._buffer[self._buffer_size :]
        self._position = len(self._buffer)

    def _upload_part(self, part_num, part_data: bytes):
        try:
            assert self._multipart_upload is not None
            self._multipart_upload.Part(part_num).upload(Body=part_data)
            self._bytes_written += len(part_data)
        except Exception as err:
            self._upload_errors.append(err)

    def _maybe_encode(self, data: AnyStr) -> bytes:
        if isinstance(data, (bytes, bytearray)):
            return data
        if self._encoding is None:
            return data.encode("utf-8")
        return data.encode(self._encoding)

    def _maybe_write(self):
        if not self._multipart_upload:
            if not self.tell():
                return
            args = self._create_object_args or {}
            args["Body"] = self._buffer
            self._s3_obj.put(**args)
            self._buffer = b""
            self._position = 0
            return

        if self.tell():
            self._start_next_part_upload()

        if self._thread_pool:
            self._thread_pool.shutdown(wait=True)
            self._thread_pool = None

        multipart_upload = self._multipart_upload
        self._multipart_upload = None

        if self._upload_errors or self._bytes_written < 1:
            multipart_upload.abort()
            self._raise_if_errored()
            return

        multipart_upload.complete(
            MultipartUpload={
                "Parts": [
                    {"ETag": part.e_tag, "PartNumber": int(part.part_number)} for part in multipart_upload.parts.all()
                ]
            }
        )

    def close(self):
        self._close()

    def _close(self, failed: bool = False):
        if self._closed:
            return

        if failed:
            if self._multipart_upload:
                self._multipart_upload.abort()
                if self._thread_pool is not None:
                    self._thread_pool.shutdown(wait=True)
            self._closed = True
            return

        self._maybe_write()

        self._closed = True



import logging
import os
from typing import Any, cast
from uuid import uuid4

import pytest
from botocore.config import Config
from botocore.exceptions import ClientError
from mypy_boto3_s3.service_resource import Bucket, ObjectSummary, ObjectVersion
from pytest_httpserver import HTTPServer

from nhs_aws_helpers import (
    dynamodb_retry_backoff,
    post_create_client,
    register_config_default,
    register_retry_handler,
    s3_build_uri,
    s3_client,
    s3_delete_keys,
    s3_get_all_keys,
    s3_list_folders,
    s3_ls,
    s3_resource,
    s3_upload_multipart_from_copy,
    transaction_cancellation_reasons,
)
from nhs_aws_helpers.common import MiB
from tests.utils import temp_env_vars


@pytest.mark.parametrize(
    ("error", "expected_reasons"),
    [
        (
            {
                "Message": "Transaction cancelled, [eeek] reasons [None, ConditionalCheckFailed, None, None, None]",
                "Code": "TransactionCanceledException",
            },
            ["None", "ConditionalCheckFailed", "None", "None", "None"],
        ),
        ({"Message": "Transaction cancelled, [eeek] specific reasons []", "Code": "TransactionCanceledException"}, []),
        ({"Message": "Transaction cancelled, [eeek] specific reasons []", "Code": "Another"}, []),
    ],
)
def test_transaction_cancellation_reasons(error, expected_reasons):
    reasons = transaction_cancellation_reasons(ClientError({"Error": error}, "TransactWriteItems"))
    assert reasons == expected_reasons


def test_ddb_retries_all_fail():
    calls: list[Any] = []

    @dynamodb_retry_backoff(max_retries=2)
    def fail_me():
        nonlocal calls
        calls.append(None)
        raise ClientError(operation_name="dummy_op", error_response={"Error": {"Code": "ThrottlingException"}})

    with pytest.raises(ClientError) as ex:
        fail_me()

    assert str(ex.value) == "An error occurred (ThrottlingException) when calling the dummy_op operation: Unknown"
    assert len(calls) == 3


def test_s3_delete_keys_all(temp_s3_bucket: Bucket):
    keys = [str(key) for key in range(1003)]

    for key in keys:
        temp_s3_bucket.put_object(Key=key, Body=f"Some data {uuid4().hex}".encode())

    deleted_keys = s3_delete_keys(keys, temp_s3_bucket.name)

    remaining_keys = list(s3_get_all_keys(temp_s3_bucket.name, ""))

    assert len(deleted_keys) == len(keys)
    assert set(deleted_keys) == set(keys)
    assert len(remaining_keys) == 0


def test_s3_delete_keys_partial(temp_s3_bucket: Bucket):
    keys = [str(key) for key in range(1003)]

    for key in keys:
        temp_s3_bucket.put_object(Key=key, Body=f"Some data {uuid4().hex}".encode())

    leaving_keys = [keys.pop(i) for i in (1002, 456, 3)]

    deleted_keys = s3_delete_keys(keys, temp_s3_bucket.name)

    remaining_keys = list(s3_get_all_keys(temp_s3_bucket.name, ""))

    assert len(deleted_keys) == len(keys)
    assert set(deleted_keys) == set(keys)
    assert len(remaining_keys) == len(leaving_keys)
    assert set(remaining_keys) == set(leaving_keys)


@pytest.mark.parametrize(
    "keys",
    [(), ("a",), ("a/b",), ("a/b", "a/c")],
)
def test_s3_get_all_keys(temp_s3_bucket: Bucket, keys: list[str]):
    for key in keys:
        temp_s3_bucket.put_object(Key=key, Body=f"Some data {uuid4().hex}".encode())

    result_keys = list(s3_get_all_keys(temp_s3_bucket.name, ""))

    assert len(result_keys) == len(keys)
    assert set(result_keys) == set(keys)


@pytest.mark.parametrize(
    "keys",
    [(), ("a",), ("a/b",), ("a/b", "a/c")],
)
def test_s3_get_all_keys_under_prefix(temp_s3_bucket: Bucket, keys: list[str]):
    expected_folder = uuid4().hex

    # Keys that shouldn't be included
    for key in ("x", "y/z"):
        temp_s3_bucket.put_object(Key=key, Body=f"Some data {uuid4().hex}".encode())

    for key in keys:
        temp_s3_bucket.put_object(Key=f"{expected_folder}/{key}", Body=f"Some data {uuid4().hex}".encode())

    result_keys = list(s3_get_all_keys(temp_s3_bucket.name, expected_folder))

    assert len(result_keys) == len(keys)
    assert set(result_keys) == {f"{expected_folder}/{key}" for key in keys}


def test_s3_ls(temp_s3_bucket: Bucket):
    expected_folder = uuid4().hex
    temp_s3_bucket.put_object(Key=f"{expected_folder}/1/a.txt", Body=f"Some data {uuid4().hex}".encode())
    temp_s3_bucket.put_object(Key=f"{expected_folder}/1/b.txt", Body=f"Some data {uuid4().hex}".encode())
    temp_s3_bucket.put_object(Key=f"{expected_folder}/2/c.txt", Body=f"Some data {uuid4().hex}".encode())

    files = cast(list[ObjectSummary], list(s3_ls(s3_build_uri(temp_s3_bucket.name, expected_folder))))

    assert len(files) == 3
    assert {f.key for f in files} == {
        f"{expected_folder}/1/a.txt",
        f"{expected_folder}/1/b.txt",
        f"{expected_folder}/2/c.txt",
    }


def test_s3_ls_versioning_on_non_versioned_bucket(temp_s3_bucket: Bucket):
    expected_folder = uuid4().hex
    temp_s3_bucket.put_object(Key=f"{expected_folder}/1/a.txt", Body=f"Some data {uuid4().hex}".encode())
    temp_s3_bucket.put_object(Key=f"{expected_folder}/1/b.txt", Body=f"Some data {uuid4().hex}".encode())
    temp_s3_bucket.put_object(Key=f"{expected_folder}/2/c.txt", Body=f"Some data {uuid4().hex}".encode())
    temp_s3_bucket.put_object(Key=f"{expected_folder}/2/c.txt", Body=f"Some new data {uuid4().hex}".encode())

    files = cast(list[ObjectVersion], list(s3_ls(s3_build_uri(temp_s3_bucket.name, expected_folder), versioning=True)))

    assert len(files) == 3
    assert {f.key for f in files} == {
        f"{expected_folder}/1/a.txt",
        f"{expected_folder}/1/b.txt",
        f"{expected_folder}/2/c.txt",
    }
    assert len({f.id for f in files}) == 1


def test_s3_ls_versioning(temp_versioned_s3_bucket: Bucket):
    expected_folder = uuid4().hex
    temp_versioned_s3_bucket.put_object(Key=f"{expected_folder}/1/a.txt", Body=f"Some data {uuid4().hex}".encode())
    temp_versioned_s3_bucket.put_object(Key=f"{expected_folder}/1/b.txt", Body=f"Some data {uuid4().hex}".encode())
    temp_versioned_s3_bucket.put_object(Key=f"{expected_folder}/2/c.txt", Body=f"Some data {uuid4().hex}".encode())
    temp_versioned_s3_bucket.put_object(Key=f"{expected_folder}/2/c.txt", Body=f"Some new data {uuid4().hex}".encode())

    files = cast(
        list[ObjectVersion], list(s3_ls(s3_build_uri(temp_versioned_s3_bucket.name, expected_folder), versioning=True))
    )

    assert len(files) == 4
    assert {f.key for f in files} == {
        f"{expected_folder}/1/a.txt",
        f"{expected_folder}/1/b.txt",
        f"{expected_folder}/2/c.txt",
    }
    assert len({f.id for f in files}) == 4


def test_s3_list_folders_root(temp_s3_bucket: Bucket):
    expected_folder = uuid4().hex

    temp_s3_bucket.put_object(Key=f"{expected_folder}/filename.txt", Body=f"Some data {uuid4().hex}".encode())

    folders = s3_list_folders(temp_s3_bucket.name, "")

    assert len(folders) == 1
    assert folders[0] == expected_folder


def test_s3_list_folders_subfolder_multiple_found(temp_s3_bucket: Bucket):
    some_body = f"Some data {uuid4().hex}".encode()
    prefix = f"{uuid4().hex}/"

    expected_folder1 = uuid4().hex
    expected_folder2 = uuid4().hex
    expected_folder3 = uuid4().hex

    temp_s3_bucket.put_object(Key=f"{prefix}{expected_folder1}/filename.txt", Body=some_body)

    temp_s3_bucket.put_object(Key=f"{prefix}{expected_folder2}/filename2.txt", Body=some_body)

    temp_s3_bucket.put_object(Key=f"{prefix}{expected_folder3}/filename2.txt", Body=some_body)

    temp_s3_bucket.put_object(Key=f"{prefix}{expected_folder3}/additional_file2.txt", Body=some_body)

    folders = s3_list_folders(temp_s3_bucket.name, prefix, page_size=2)

    assert len(folders) == 3
    assert f"{prefix}{expected_folder1}" in folders
    assert f"{prefix}{expected_folder2}" in folders
    assert f"{prefix}{expected_folder3}" in folders


@pytest.mark.parametrize(
    ("is_resource", "expected_exception_message"),
    [
        (
            True,
            "An error occurred (503) when calling the PutObject operation"
            " (reached max retries: 1): Service Unavailable",
        ),
        (
            False,
            "An error occurred (500) when calling the ListBuckets operation"
            " (reached max retries: 1): Internal Server Error",
        ),
    ],
)
def test_s3_retries(
    temp_s3_bucket: Bucket,
    httpserver: HTTPServer,
    is_resource: bool,
    expected_exception_message: str,
):
    server_port = httpserver.port

    expected_folder = uuid4().hex

    key = f"{expected_folder}/filename.txt"
    httpserver.expect_request(f"/{temp_s3_bucket.name}/{key}").respond_with_json({}, status=503)

    config = Config(
        connect_timeout=float(os.environ.get("BOTO_CONNECT_TIMEOUT", "1")),
        read_timeout=float(os.environ.get("BOTO_READ_TIMEOUT", "1")),
        max_pool_connections=int(os.environ.get("BOTO_MAX_POOL_CONNECTIONS", "10")),
        retries={
            "mode": os.environ.get("BOTO_RETRIES_MODE", "standard"),  # type: ignore[typeddict-item]
            "total_max_attempts": int(os.environ.get("BOTO_RETRIES_TOTAL_MAX_ATTEMPTS", "2")),
        },
    )

    def _post_create_aws(boto_module: str, _: str, client):
        if boto_module != "s3":
            return
        register_retry_handler(client)

    register_config_default("s3", config)
    post_create_client(_post_create_aws)

    with (  # noqa: PT012
        pytest.raises(ClientError) as ex,
        temp_env_vars(
            # Point any new boto resources or clients at our fake endpoint
            AWS_ENDPOINT_URL=f"http://localhost:{server_port}",
        ),
    ):
        original_log_level = logging.getLogger("werkzeug").level
        try:
            # Avoid werkzeug logs cluttering our logs
            logging.getLogger("werkzeug").setLevel(logging.WARNING)
            if is_resource:
                s3_resource(config=config).Bucket(temp_s3_bucket.name).put_object(
                    Key=f"{key}", Body=f"Some data {uuid4().hex}".encode()
                )
            else:
                s3_client(config=config).list_buckets()

        finally:
            logging.getLogger("werkzeug").setLevel(original_log_level)

    assert str(ex.value) == expected_exception_message


# pylint: disable=too-many-locals
def test_s3_upload_multipart_from_copy(temp_s3_bucket: Bucket):
    uuid = uuid4().hex
    num_parts = 3
    ten_mb = b"ABCDEFGHIJ" * MiB
    chunk_size = 5 * MiB
    chunk_data = ten_mb[:chunk_size]

    part_keys = []
    for i in range(1, num_parts + 1):
        part_key = f"{uuid}_{i}"
        temp_s3_bucket.Object(part_key).put(Body=chunk_data)
        part_keys.append(part_key)

    target_object = temp_s3_bucket.Object(uuid)

    s3_upload_multipart_from_copy(target_object, part_keys)

    assert target_object.get()["ContentLength"] == num_parts * chunk_size


def test_s3_upload_multipart_from_copy_one_part_too_small(temp_s3_bucket: Bucket):
    uuid = uuid4().hex
    num_parts = 3
    ten_mb = b"ABCDEFGHIJ" * MiB

    part_keys = []
    for i in range(1, num_parts + 1):
        part_key = f"{uuid}_{i}"
        # 4, 8, 12 MB - 4 is too small
        temp_s3_bucket.Object(part_key).put(Body=ten_mb[: i * 4 * MiB])
        part_keys.append(part_key)

    target_object = temp_s3_bucket.Object(uuid)

    with pytest.raises(ClientError) as client_error:
        s3_upload_multipart_from_copy(target_object, part_keys)

    assert client_error.value.response["Error"]["Code"] == "EntityTooSmall"

    with pytest.raises(ClientError) as client_error:
        target_object.get()

    assert client_error.value.response["Error"]["Code"] == "NoSuchKey"


# pylint: disable=too-many-locals
def test_s3_upload_multipart_from_copy_missing_part_data(temp_s3_bucket: Bucket):
    uuid = uuid4().hex
    num_parts = 3
    ten_mb = b"ABCDEFGHIJ" * MiB
    chunk_size = 5 * MiB
    chunk_data = ten_mb[:chunk_size]

    part_keys = []
    for i in range(1, num_parts + 1):
        part_key = f"{uuid}_{i}"
        if i != 2:  # Don't store part 2
            temp_s3_bucket.Object(part_key).put(Body=chunk_data)
        part_keys.append(part_key)

    target_object = temp_s3_bucket.Object(uuid)

    with pytest.raises(IOError, match=f"Failed to multipart_upload {{'{uuid}_2'}} after 5 attempts"):
        s3_upload_multipart_from_copy(target_object, part_keys)

    with pytest.raises(ClientError) as client_error:
        target_object.get()

    assert client_error.value.response["Error"]["Code"] == "NoSuchKey"
